/***************************************************************************
 *   Copyright (C) 2008 by Ingo Kresse                                     *
 *   kresse@in.tum.de                                                      *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/



/*! This program will connect to an X-IST data glove and
record the data to stdout
 */


#include <fcntl.h> // low-level serial commands
#include <termios.h> // serial commands
#include <signal.h> // SIGINT signal handling
#include <stdio.h> // printf
#include <time.h>  // for timestamps

// VARIABLES

int shouldExit=0;

int fd;
struct termios oldSerialOptions;

// read buffer
#define BUFFER_LEN 256

char buffer[BUFFER_LEN+1];
int raw[24]; //!< raw data values

int firmware;
char firmwareInfo[BUFFER_LEN+1],
     hardwareInfo[BUFFER_LEN+1],
     productInfo[BUFFER_LEN+1];


// FUNCTIONS

int Write(char data); //!< write a single character
int Read(char *buffer, int numBytes); //!< read characters into buffer

//! comm functions
int ReadValues(); //!< request and read current values into buffer
int parseValues(char *buffer); //!< parse data values into raw[]

int readInfo(); //!< read version numbers

int getTime(char *dst); //!< helper function for logging

//! Ctrl-C handler
void catchsignal(int signo)
{
  shouldExit=1;
}


int main(int argc, char *argv[])
{
  if(argc < 2) {
    fprintf(stderr, "usage: glovelog <device>\n");
    return 0;
  }

  // install Ctrl-C handler
  struct sigaction act;
  act.sa_handler = catchsignal;
  act.sa_flags = 0;
  if ((sigemptyset(&act.sa_mask) == -1) ||
       (sigaction(SIGINT, &act, NULL) == -1)) {
    printf("Failed to set SIGINT to handle Ctrl-C, oh well\n");
  }

  openConnection(argv[1]);

  while(shouldExit == 0) {
    if(ReadValues() > 0) {
      parseValues(buffer);

      // system time
      char c[40];
      getTime(c);
      fprintf(stdout, "%s",c);
      int i;
      for(i=0; i < 24; i++) {
        fprintf(stdout, " %d", raw[i]);
      }
      fprintf(stdout, "\n");
    }
  }

  fprintf(stderr, "## exiting graciously\n");
  return 0;
}


int getTime(char *dst)
{
  char buf[40];

  struct timeval rawtime;
  gettimeofday(&rawtime, 0);

//  struct tm *calendartime;
//  time_t tsecs = (time_t) (rawtime.tv_sec);  
//  calendartime = localtime(&tsecs);

//  strftime(buf, 40, "%H:%M:%S", calendartime);
//  sprintf(dst, "%s.%03d", buf, (rawtime.tv_usec/1000));

  sprintf(dst, "%d.%03d", rawtime.tv_sec, rawtime.tv_usec/1000);
}


/*! Interface options is 1000000 Baud, 8N1, no flow control.
 * we need raw data
 */
int openConnection(char *port)
{

  fd = -1;

  int rate = B1000000;

  // open the serial port

  //fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);
  fd = open(port, O_RDWR | O_NOCTTY);
  if ( fd<0 )
  {
    fprintf(stderr, "Could not open serial device %s\n",port);
    return;
  }
  fcntl(fd, F_SETFL, FNDELAY); // read calls will NOT block

  // save the current io settings
  tcgetattr(fd, &oldSerialOptions);

  struct termios options;

  options=oldSerialOptions;
  options.c_cflag &= ~(PARENB | CSTOPB | CSIZE);
  options.c_cflag |= CS8;	

  options.c_cflag &= ~CRTSCTS; // disable flow control
  options.c_cflag |= CLOCAL | CREAD; // process control stuff

  options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // raw input mode
  options.c_iflag &= ~(IXON | IXOFF | IXANY); // disable software flow control

  options.c_oflag &= ~OPOST; // disable output postprocessing (= raw output)

  if (cfsetispeed(&options, rate) < 0 || cfsetospeed(&options, rate) < 0)
  {
    fprintf(stderr,"Failed to set serial baud rate: %d\n", rate);
    tcsetattr(fd, TCSANOW, &oldSerialOptions);	
    close(fd);
    fd = -1;
    return;
  }

  // activate new settings
  tcflush(fd, TCIFLUSH);
  tcsetattr(fd, TCSANOW, &options);
  tcflush(fd, TCIFLUSH);

  if(readInfo() < 1) {
    // if it really failed give up and disable the driver
    tcsetattr(fd, TCSANOW, &oldSerialOptions);	
    fd = -1;
  }

  fcntl(fd,F_SETFL, 0); // read calls will block	
}

int readInfo()
{
  Write(0x11); // get firmware version
  usleep(50000);
  int l1 = Read(firmwareInfo, 40);
  firmwareInfo[l1] = '\0'; // just to make shure

  firmware = (firmwareInfo[0]-'0')*100
           + (firmwareInfo[2]-'0')*10
           + (firmwareInfo[4]-'0');

  Write(0x12); // get hardware version
  usleep(50000);
  int l2 = Read(hardwareInfo, 40);
  hardwareInfo[l2] = '\0'; // just to make shure

  Write(0x13); // get hardware version
  usleep(50000);
  int l3 = Read(productInfo, 40);
  productInfo[l3] = '\0'; // just to make shure

  // delay here so data has arrived at serial port so we can flush it
  usleep(100000);
  tcflush(fd, TCIFLUSH);


  if (l1 < 3 || l2 < 3 || l3 < 3) // got an error
  {
    return -1;
  }

  fprintf(stderr, "## Firmware Version: %s\n## Hardware Version: %s\n## Device Name: %s\n", firmwareInfo, hardwareInfo, productInfo);
}

int Write(char cmd)
{
  if (fd < 0)
    return -1;
  if(write(fd, &cmd, 1) < 1)
    return -1;
//  fprintf(stderr, "## written 0x%02x\n",cmd);

  return 1;
}


int Read(char *buffer, int len)
{
  int l;
  if (fd < 0)
    return -1;
  l=read(fd, buffer, len );
//  fprintf(stderr, "## read %d chars\n", l);
  return (l<0) ? 0 : l;
}


int ReadValues()
{
  if (fd < 0)
    return -1;

  Write(0x35);
  usleep(10000);
  int len = 0;
  len = Read(buffer, 36-len);

//  fprintf(stderr,"got %d bytes\n", len);	

  // TODO: add more error handling: flush buffers if everthing else fails
  while(len < 36)
  {
    len += Read(buffer+len, 36-len);

    fprintf(stderr,"## Error getting data (got %d / 36)\n", len);
    //return -1;
  }

  return 1;
}

int parseValues(char *buffer)
{
  unsigned char *buf;
  buf = (unsigned char *) buffer;

  raw[0] = buf[0] + (buf[1]/16)*256;
  raw[1] = buf[2] + (buf[1]%16)*256;
  raw[2] = buf[3] + (buf[4]/16)*256;
  raw[3] = buf[5] + (buf[4]%16)*256;
  raw[4] = buf[6] + (buf[7]/16)*256;
  raw[5] = buf[8] + (buf[7]%16)*256;
  raw[6] = buf[9] + (buf[10]/16)*256;
  raw[7] = buf[11] + (buf[10]%16)*256;
  raw[8] = buf[12] + (buf[13]/16)*256;
  raw[9] = buf[14] + (buf[13]%16)*256;
  raw[10] = buf[15] + (buf[16]/16)*256;
  raw[11] = buf[17] + (buf[16]%16)*256;
  raw[12] = buf[18] + (buf[19]/16)*256;
  raw[13] = buf[20] + (buf[19]%16)*256;
  raw[14] = buf[21] + (buf[22]/16)*256;
  raw[15] = buf[23] + (buf[22]%16)*256;
  raw[16] = buf[24] + (buf[25]/16)*256;
  raw[17] = buf[26] + (buf[25]%16)*256;
  raw[18] = buf[27] + (buf[28]/16)*256;
  raw[19] = buf[29] + (buf[28]%16)*256;
  if (firmware > 26)
  {
    raw[20] = buf[30] + (buf[31]/16)*256;
    raw[21] = buf[32] + (buf[31]%16)*256;
    raw[22] = buf[33] + (buf[34]/16)*256;
    raw[23] = buf[35] + (buf[34]%16)*256;
    return 24;
  }
  else
    return 20;
}
