/** This program watches the sensor readings of an X-IST dataglove.
 *  The maximum and minimum raw values are recorded and written
 *  to stdout when the program exits.
 */

#include <signal.h> // SIGINT signal handling

#include <iostream>
#include <fstream>

#include <math.h>
#include <GL/glut.h>

#include "Trackball.h"

#include "args.h"

#include <libplayerc++/playerc++.h>

using namespace PlayerCc;

using namespace std;

// Some <math.h> files do not define PI...
#ifndef PI
#define PI 3.14159265358979323846
#endif

// raw ranges
int minRawValue[24] = {4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096};
int maxRawValue[24] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};


void readCalibration(istream &stream)
{
	char id='\0';
	bool complete=false;
	while(stream && !complete) {
		stream >> id;
		switch(id) {
			case 'L':
				for(int i=0; i < 24; i++)
					stream >> minRawValue[i];
				break;
			case 'U':
				for(int i=0; i < 24; i++)
					stream >> maxRawValue[i];
				break;
			case ';':
			default: // this would mean an error!
				complete=true;
				break;
		}
	}
}

void writeCalibration(ostream &stream)
{
	stream << "L";
	for(int i=0; i < 24; i++)
		stream << " " << minRawValue[i];
	stream << endl << "U";
	for(int i=0; i < 24; i++)
		stream << " " << maxRawValue[i];
	stream << endl << ";" << endl;
}


PlayerClient *pclient;
AioProxy *aio;

// ring buffer for acceleration sensors
#define BUFFER_SIZE 32
int cBuffer[2][BUFFER_SIZE];
int cIndex=0;


void idleFunc()
{
	pclient->Read();

	for(unsigned int i=0; i < aio->GetCount(); i++) {
		if(i == 20 || i == 21) {
			int index = i-20;
			cBuffer[index][cIndex] = (int) aio->GetVoltage(i);
			cIndex = (cIndex + index) % BUFFER_SIZE;
			
			// determine signal energy
			int energy = 0;
			int e = 0;

			if(cIndex != 0)
			{
				e = cBuffer[index][0] - cBuffer[index][BUFFER_SIZE-1];
				energy = e*e;
			}

			for(int j=1; j < BUFFER_SIZE; j++) {
				if(j != cIndex) {
					e = cBuffer[index][j] - cBuffer[index][j-1];
					energy += e*e;
				}
			}

			if(energy > 300) {
				continue;
			}
		}
		// plain maximum/minimum recording
		maxRawValue[i] = ((int) aio->GetVoltage(i) > maxRawValue[i]) ? (int) aio->GetVoltage(i) : maxRawValue[i];
		minRawValue[i] = ((int) aio->GetVoltage(i) < minRawValue[i]) ? (int) aio->GetVoltage(i) : minRawValue[i];
	}
}


//! Ctrl-C handler
void catchsignal(int signo)
{
	writeCalibration(cout);
	exit(0);
}


int main(int argc, char *argv[])
{
	// install Ctrl-C handler
	struct sigaction act;
	act.sa_handler = catchsignal;
	act.sa_flags = 0;
	if ((sigemptyset(&act.sa_mask) == -1) ||
	    (sigaction(SIGINT, &act, NULL) == -1)) {
		printf("Failed to set SIGINT to handle Ctrl-C, oh well\n");
	}

	parse_args(argc, argv);
	
	pclient = new PlayerClient(gHostname, gPort);
	pclient->SetDataMode(PLAYER_DATAMODE_PUSH); // doesn't help :-(
	pclient->SetReplaceRule(false, PLAYER_MSGTYPE_DATA, -1, -1);
	
	aio = new AioProxy(pclient, gIndex);
	aio->SetReplaceRule(PLAYER_MSGTYPE_DATA,-1);

	if(gCalibrationFile != "") {
		cerr << "loading calibration file " << gCalibrationFile << endl;
		ifstream file(gCalibrationFile.c_str());
		readCalibration(file);
		file.close();
	}

	for(int i=0; i < BUFFER_SIZE; i++) {
		cBuffer[0][i] = 0;
		cBuffer[1][i] = 0;
	}

	// main loop
	while(1) {
		idleFunc();
	}

	return 0;
}
