/** This program displays the sensor readings of an X-IST dataglove
 *  in an intuitive way. It reads the data from the player plugin and connects
 *  to its aio interface.
 *
 *  During runtime the maximum and minimum raw values are recorded which are
 *  used for calibration.
 *
 *  The missing thumb joint is fixed to 30 degrees.
 *
 *  Usage: LMB-drag        : rotates
 *         Shift-LMB-drag  : zooms
 *
 */


#include <stdlib.h>
#include <stdio.h>

#include <iostream>
#include <fstream>

#include <math.h>
#include <GL/glut.h>

#include "Trackball.h"

#include "args.h"

#include <libplayerc++/playerc++.h>

using namespace PlayerCc;

using namespace std;

// Some <math.h> files do not define PI...
#ifndef PI
#define PI 3.14159265358979323846
#endif


float fingerRoot[][3] = {{-2,20,-20},{99,31,0},{96,6,0},{89,-11,0},{81,-29,0}};
float fingerLength[][3] = {{78,26,30},{49,28,21},{54,32,23},{48,31,24},{38,22,21}};

// configuration angles
float fingerJointAngle[][3] = {{45, 45, 0},{0, 0, 0},{0, 90, 0},{0, 90, 0},{0, 0, 0}};
float thumbExtraJoint = 90.0;

// target ranges
//TODO: adjust these angles
float minAngle[][3] = {{0,0,0},{0,0,0},{0,0,0},{0,0,0},{0,0,0}};
float maxAngle[][3] = {{90,90,90},{90,90,90},{90,90,90},{90,90,90},{90,90,90}};

float extraMin = 0;
float extraMax = 160;


float force[5] = {0, 0, 0, 0, 0};  // finger tip forces
float orientation[3] = {0,0,0};    // tilt sensor

// raw ranges
int minRawValue[24] = {4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096,4096};
int maxRawValue[24] = {0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0,0};



Trackball *trackball=0; // virtual trackball for scene rotation

// set up lighting and material
void setupLighting() {
	glEnable(GL_LIGHTING);
	glEnable(GL_LIGHT0);
	
	GLfloat light_pos[] = {35.0, -35.0, -35.0, 1.0};
	GLfloat light_Ka[] = {0.0, 0.0, 0.0, 1.0};
	GLfloat light_Kd[] = {1.0, 1.0, 1.0, 1.0};
	GLfloat light_Ks[] = {1.0, 1.0, 1.0, 1.0};
		
	glLightfv(GL_LIGHT0, GL_POSITION, light_pos);
	glLightfv(GL_LIGHT0, GL_AMBIENT, light_Ka);
	glLightfv(GL_LIGHT0, GL_DIFFUSE, light_Kd);
	glLightfv(GL_LIGHT0, GL_SPECULAR, light_Ks);
	
	glEnable(GL_COLOR_MATERIAL);
	glColorMaterial(GL_FRONT_AND_BACK, GL_DIFFUSE); //glColor changes the diffuse color
	
	GLfloat material_Ka[] = {0.12, 0.12, 0.12, 1.0};
	GLfloat material_Kd[] = {0.5, 0.5, 0.5, 1.0};
	GLfloat material_Ks[] = {0.3, 0.3, 0.3, 1.0};
	GLfloat material_Ke[] = {0.0, 0.0, 0.0, 1.0};
	GLfloat material_Se = 10;
	
	glMaterialfv(GL_FRONT, GL_AMBIENT, material_Ka);
	glMaterialfv(GL_FRONT, GL_DIFFUSE, material_Kd);
	glMaterialfv(GL_FRONT, GL_SPECULAR, material_Ks);
	glMaterialfv(GL_FRONT, GL_EMISSION, material_Ke);
	glMaterialf(GL_FRONT, GL_SHININESS, material_Se);
}


void init()
{
	glClearColor(0.4, 0.4, 0.4, 0.0);

	// set up view
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	gluPerspective(65.654, 4.0/3.0, 1.0, 10000.0);
	// put a canonical camera on the projection stack
	gluLookAt(0.0, 0.0, 0.0,
	          0.0, 0.0, 1.0,
	          0.0, 1.0, 0.0);


	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	// here we adjust our view
	glTranslatef(-50.0, 0.0, 250.0);
	glRotatef(90, 1, 0, 0);

	setupLighting();
	
	glClearDepth(1.0);
	glEnable(GL_DEPTH_TEST);
	glDepthFunc(GL_LEQUAL);
	
	glEnable(GL_CULL_FACE);
	glCullFace(GL_BACK);
	
	trackball = new Trackball(640, 480, 30);	
	glPushMatrix();
}



// ----------------------------------------------------------------------
void drawFunc(void)
{
	GLUquadric* quad = gluNewQuadric();  // create helper object

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	trackball->rotate();

//	glColor3f(1.0, 1.0, 1.0);
//	glPushMatrix();
//	glTranslatef(35, -35, -35);
//	gluSphere(quad, 9.0, 10, 5);  // light position sphere
//	glPopMatrix();

	// coordinate frame
	glDisable(GL_LIGHTING);
	glBegin(GL_LINES);

	glColor3f(1, 0, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(30, 0, 0);

	glColor3f(0, 1, 0);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 30, 0);

	glColor3f(0, 0, 1);
	glVertex3f(0, 0, 0);
	glVertex3f(0, 0, 30);

	glEnd();
	glEnable(GL_LIGHTING);


	// hand rotation
//	glRotatef(orientation[1], 0, -1, 0);	
//	glRotatef(orientation[0], 1, 0, 0);	

	// rotation axis is the cross product between "palm vector" and down vector
	float rotx = orientation[1];
	float roty = -orientation[0];
	float rotz = 0;

	orientation[2] = (orientation[2] < 0) ? 0 : orientation[2];
	float angle = (orientation[2] > 1) ? PI / 2.0 : acos(orientation[2]);

//	printf("o: [%f, %f, %f]  angle: %f\n", orientation[0], orientation[1], orientation[2], angle);

	glRotatef(angle*180.0/PI, rotx, roty, rotz);	
	glColor3f(1.0, 1.0, 0.0);

	// draw palm
	glBegin(GL_POLYGON);
	glNormal3f(0.0, 0.0, -1.0);
	for(int i = 0; i < 5; i++) {
		glVertex3f(fingerRoot[i][0], fingerRoot[i][1], fingerRoot[i][2]-7);
	}
	glVertex3f(fingerRoot[0][0], fingerRoot[0][1]-30, fingerRoot[0][2]-7);
	glEnd();

	glBegin(GL_POLYGON);
	glNormal3f(0.0, 0.0, 1.0);
	glVertex3f(fingerRoot[0][0], fingerRoot[0][1]-30, fingerRoot[0][2]+7);
	for(int i = 4; i >= 0; i--) {
		glVertex3f(fingerRoot[i][0], fingerRoot[i][1], fingerRoot[i][2]+7);
	}
	glEnd();

	glBegin(GL_QUAD_STRIP);
	for(int i = 0; i < 5; i++) {
		glVertex3f(fingerRoot[i][0], fingerRoot[i][1], fingerRoot[i][2]-7);
		glVertex3f(fingerRoot[i][0], fingerRoot[i][1], fingerRoot[i][2]+7);
	}
	glVertex3f(fingerRoot[0][0], fingerRoot[0][1]-30, fingerRoot[0][2]-7);
	glVertex3f(fingerRoot[0][0], fingerRoot[0][1]-30, fingerRoot[0][2]+7);
	glVertex3f(fingerRoot[0][0], fingerRoot[0][1], fingerRoot[0][2]-7);
	glVertex3f(fingerRoot[0][0], fingerRoot[0][1], fingerRoot[0][2]+7);
	glEnd();

	//TODO: adjust position of thumb root (z-value)

	// draw fingers
	for(int i = 0; i < 5; i++) {
		glPushMatrix();
		glTranslatef(fingerRoot[i][0], fingerRoot[i][1], fingerRoot[i][2]);
		gluSphere(quad, 9.0, 10, 5);  // draw hand roots
		glRotatef(90, 0, 1, 0);
		for(int j = 0; j < 3; j++) {
			// do configuration-dependent rotations
			if(i==0 && j == 0) { // thumb root joint
				glRotatef(90, 0, 0, 1);
				glRotatef(-thumbExtraJoint, 0, 0, 1);
				glRotatef(180, 0, 0, 1);
				glRotatef(-fingerJointAngle[i][j], 0, 1, 0);

				glRotatef(thumbExtraJoint/2.0, 0, 0, 1); // the thumb is not turned that strongly around its axis ...	

			} else {
				glRotatef(fingerJointAngle[i][j], 0, 1, 0);
			}

			gluCylinder(quad, 5.0, 5.0, fingerLength[i][j], 17, 1);
			glTranslatef(0.0, 0.0, fingerLength[i][j]);
			if(j == 2)
				glColor3f(1.0, 1.0 - force[i], 0.0);
			gluSphere(quad, 7.0, 15, 10);

			glColor3f(1.0, 1.0, 0.0);
		}
		glPopMatrix();
	}

	/* Swap buffers */
	glutSwapBuffers();
}

void readCalibration(istream &stream)
{
	char id='\0';
	bool complete=false;
	while(stream && !complete) {
		stream >> id;
		switch(id) {
			case 'L':
				for(int i=0; i < 24; i++)
					stream >> minRawValue[i];
				break;
			case 'U':
				for(int i=0; i < 24; i++)
					stream >> maxRawValue[i];
				break;
			case ';':
			default: // this would mean an error!
				complete=true;
				break;
		}
	}
}

void writeCalibration(ostream &stream)
{
	stream << "L";
	for(int i=0; i < 24; i++)
		stream << " " << minRawValue[i];
	stream << endl << "U";
	for(int i=0; i < 24; i++)
		stream << " " << maxRawValue[i];
	stream << endl << ";" << endl;
}


// This function is called every time a key is hit.
void keyboardFunc(unsigned char ch, int x, int y)
{
	switch (ch) {
		case 27:             // ESC
			exit(0);
			break;
		case 'c':
			writeCalibration(cout);
			break;
	}
}

void mouseFunc(int button, int state, int x, int y)
{
	trackball->mouse(button, state, x, 480-y);	
}

void motionFunc(int x, int y)
{
	trackball->motion(x, 480-y);
	glutPostRedisplay();
}

PlayerClient *pclient;
AioProxy *aio;

// circular buffer for acceleration sensors
#define BUFFER_SIZE 32
//int bufferX1[1000];
int cBuffer[2][BUFFER_SIZE];
//int bufferX2[1000];
int cIndex=0;

// readings get scaled to [0..1]
float getNormalized(int index) {
	return (aio->GetVoltage(index) - minRawValue[index]) / (maxRawValue[index] - minRawValue[index]);
}

// readings get scaled to [0..1]
float getFilteredNormalized(int index)
{
	float filter[] = {1.0, 0.5, 0.25};
	float filter_sum = 1.75;
	int filter_len = 3;

	float mean=0;
	int id = index - 20;

	for(int i=0; i < filter_len; i++)
		mean += filter[i]*cBuffer[id][(cIndex+BUFFER_SIZE-i-1) % BUFFER_SIZE];

	mean /= filter_sum;
	
	return (mean - minRawValue[index]) / (maxRawValue[index] - minRawValue[index]);
}


// readings get rescaled to target range
float getRescaled(int index, int finger, int joint)
{
	return (1.0 - getNormalized(index)) * (maxAngle[finger][joint] - minAngle[finger][joint]) +
	minAngle[finger][joint];
}



void idleFunc()
{
	// wait for player data
	pclient->Read();

	for(unsigned int i=0; i < aio->GetCount(); i++) {
		if(i == 20 || i == 21) {
			int index = i- 20;
			cBuffer[index][cIndex] = (int) aio->GetVoltage(i);
			cIndex = (cIndex + index) % BUFFER_SIZE;

			// determine signal energy
			int energy = 0;
			int e = 0;

			if(cIndex != 0)
			{
				e = cBuffer[index][0] - cBuffer[index][BUFFER_SIZE-1];
				energy = e*e;
			}

			for(int j=1; j < BUFFER_SIZE; j++) {
			
				if(j != cIndex) {
					e = cBuffer[index][j] - cBuffer[index][j-1];
					energy += e*e;
				}
			}

//			if(cIndex % 16 == 0 && index == 0)
//				printf("energy(%d): %d\n", index, energy);

			if(energy > 300) {
				continue;
			}
		}
		// plain maximum/minimum recording
		maxRawValue[i] = ((int) aio->GetVoltage(i) > maxRawValue[i]) ? (int) aio->GetVoltage(i) : maxRawValue[i];
		minRawValue[i] = ((int) aio->GetVoltage(i) < minRawValue[i]) ? (int) aio->GetVoltage(i) : minRawValue[i];
	}

	// thumb
	fingerJointAngle[0][0] = 30;  // we can't detect this one :-(
	thumbExtraJoint = (1.0 - getNormalized(2)) * (extraMax - extraMin) + extraMin;

	fingerJointAngle[0][1] = getRescaled(0, 0, 1);
	fingerJointAngle[0][2] = getRescaled(1, 0, 2);

	// index finger
	fingerJointAngle[1][0] = getRescaled(5, 1, 0);
	fingerJointAngle[1][1] = getRescaled(4, 1, 1);
	fingerJointAngle[1][2] = getRescaled(3, 1, 2);

	// middle finger
	fingerJointAngle[2][0] = getRescaled(8, 2, 0);
	fingerJointAngle[2][1] = getRescaled(7, 2, 1);
	fingerJointAngle[2][2] = getRescaled(6, 2, 2);

	// ring finger
	fingerJointAngle[3][0] = getRescaled(11, 3, 0);
	fingerJointAngle[3][1] = getRescaled(10, 3, 1);
	fingerJointAngle[3][2] = getRescaled(9, 3, 2);

	// pinkie finger
	fingerJointAngle[4][0] = getRescaled(14, 4, 0);
	fingerJointAngle[4][1] = getRescaled(13, 4, 1);
	fingerJointAngle[4][2] = getRescaled(12, 4, 2);

	// finger tip forces
	force[0] = getNormalized(15);
	force[1] = getNormalized(16);
	force[2] = getNormalized(17);
	force[3] = getNormalized(18);
	force[4] = getNormalized(19);

	// tilt readings
	// this is the projection of the gravity vector into the palm plane (assuming a length of 1)
	orientation[0] = getFilteredNormalized(21) * 2 - 1;
	orientation[1] = getFilteredNormalized(20) * 2 - 1;
	float lpalm = sqrt(orientation[0]*orientation[0] + orientation[1]*orientation[1]);
	lpalm = (lpalm < 1) ? lpalm : 1;
	orientation[2] = sqrt(1 - lpalm);

	glutPostRedisplay();
}


int main(int argc, char *argv[])
{
	// Initialize GLUT
	glutInitWindowPosition(10, 10);
	glutInitWindowSize(640, 480);
	glutInit(&argc, argv);
	glutInitDisplayMode(GLUT_DOUBLE | GLUT_RGB | GLUT_DEPTH);
	glutCreateWindow("OpenGL dataglove viewer");

	glutDisplayFunc(drawFunc);
	glutMouseFunc(mouseFunc);
	glutMotionFunc(motionFunc);
	glutKeyboardFunc(keyboardFunc);
	glutIdleFunc(idleFunc);

	init();

	parse_args(argc, argv);

	pclient = new PlayerClient(gHostname, gPort);
	pclient->SetDataMode(PLAYER_DATAMODE_PUSH); // doesn't help :-(
	pclient->SetReplaceRule(false, PLAYER_MSGTYPE_DATA, -1, -1);

	aio = new AioProxy(pclient, gIndex);
	aio->SetReplaceRule(PLAYER_MSGTYPE_DATA,-1);

	if(gCalibrationFile != "") {
		cerr << "loading calibration file " << gCalibrationFile << endl;
		ifstream file(gCalibrationFile.c_str());
		readCalibration(file);
		file.close();
	}

	for(int i=0; i < BUFFER_SIZE; i++) {
		cBuffer[0][i] = 0;
		cBuffer[1][i] = 0;
	}

	glutMainLoop();
	return 0;
}
