/** This program displays the sensor readings of an X-IST dataglove
 *  in an intuitive way. It reads the data from the player plugin and connects
 *  to its aio interface.
 *
 *  During runtime the maximum and minimum raw values are recorded which are
 *  used for calibration.
 *
 *  The missing thumb joint is fixed to 30 degrees.
 *
 *  Usage: LMB-drag        : rotates
 *         Shift-LMB-drag  : zooms
 *
 */

#include <signal.h> // SIGINT signal handling

#include <iostream>
#include <fstream>

#include <libplayerc++/playerc++.h>
#include <unistd.h>

using namespace PlayerCc;
using namespace std;

string  gHostname(PLAYER_HOSTNAME);
uint    gPort=PLAYER_PORTNUM;
uint    gIndex=0;
bool    gStart=false;
bool    gStop=false;
bool    gRewind=false;
bool    gWait=false;
string  gFileName="";


void print_usage(int argc, char** argv)
{
  cerr << "USAGE:  " << *argv << " [options]" << endl;
  cerr << endl << "Where [options] can be:" << endl;
  cerr << endl << "  -h <hostname>  : hostname to connect to (default: "
       << PlayerCc::PLAYER_HOSTNAME << ")";
  cerr << endl << "  -p <port>      : port where Player will listen (default: "
       << PlayerCc::PLAYER_PORTNUM << ")";
  cerr << endl << "  -i <index>     : aio device index (default: 0)";
  cerr << endl << "  -s             : start logging";
  cerr << endl << "  -t             : stop logging";
  cerr << endl << "  -r             : rewind";
  cerr << endl << "  -w             : wait until logging is finished";
  cerr << endl << "  -f <filename>  : set filename";
  cerr << endl << "  -?             : display this help";
  cerr << endl << "returns the current state (active/inactive = 1/0)"
       << endl;
} // end print_usage


int parse_args(int argc, char** argv)
{
  // set the flags
  const char* optflags = "h:p:i:strwf:?";
  int ch;

  // use getopt to parse the flags
  while(-1 != (ch = getopt(argc, argv, optflags)))
  {
    switch(ch)
    {
      // case values must match long_options
      case 'h': // hostname
          gHostname = optarg;
          break;
      case 'p': // port
          gPort = atoi(optarg);
          break;
      case 'i': // index
          gIndex = atoi(optarg);
          break;
      case 's': // start
          gStart=true;
          break;
      case 't': // stop
          gStop=true;
          break;
      case 'r': // rewind
          gRewind=true;
          break;
      case 'w': // wait
          gWait=true;
          break;
      case 'f': // filename
          gFileName = optarg;
          break;
      case '?': // help
      case ':':
      default:  // unknown
        print_usage(argc, argv);
        exit (-1);
    }
  }
  return (0);
} // end parse_args

#if 0
void setState(LogProxy *log, int aState)
{
  if(log->GetType() == PLAYER_LOG_TYPE_READ)
    log->SetReadState(aState);
  else
    log->SetWriteState(aState);
}
#endif

int main(int argc, char *argv[])
{
  parse_args(argc, argv);

  cerr << "connecting to " << gHostname << ":" << gPort << endl;

  PlayerClient *pclient = new PlayerClient(gHostname, gPort);
  LogProxy *log = new LogProxy(pclient, gIndex);

  log->QueryState();         // get interface info
  int type = log->GetType(); // will be either PLAYER_LOG_TYPE_READ or PLAYER_LOG_TYPE_WRITE

  cerr << "connected to log" << ((type == PLAYER_LOG_TYPE_READ) ? "reader " : "writer ") << gIndex << endl;

  const char *action = (type == PLAYER_LOG_TYPE_READ) ? "playback " : "logging ";

  if(gStop) {
    log->SetState(0);
  }

  if(gFileName != "") {
    if(type == PLAYER_LOG_TYPE_WRITE) {
      log->QueryState();
      int state = log->GetState();

      if(state == 1) log->SetState(0);  // we must stop the logwriter first

      cerr << "changing filename " << gIndex << " to " << gFileName << endl;
      log->SetFilename(gFileName);

      if(state == 1) log->SetState(1);  // now we can activate the logwriter again
    } else {
      cerr << "logreader does not (yet) support changing the filename" << endl;
    }
  }

  if(gRewind) {
    log->Rewind();
  }

  if(gStart) {
    log->SetState(1);
  }

  if(gWait) {
    int active = log->GetState();
    int seenActive = active;

    if(active)
      cerr << action << gIndex << " in progress" << endl;
    else
      cerr << action << gIndex << " inactive" << endl;
    while(!seenActive || active) {
      log->QueryState();
      active = log->GetState();

      if(seenActive == 0 && active == 1)
        cerr << action << gIndex << " started" << endl;

      seenActive |= active;
    }
    cerr << action << gIndex << " stopped" << endl;
  }

  return log->GetState();
}
