/*
 *  Player - One Hell of a Robot Server
 *  Copyright (C) 2007 Ingo Kresse kresse \/at\/ in.tum.de
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 */

/** @ingroup drivers */
/** @{ */
/** @defgroup driver_dataglove dataglove
 * @brief X-IST Dataglove driver

This driver interfaces the dataglove from X-IST, equipped with force-, bend- and tilt-sensors. The number of output channels is determined automatically by the firmware version, which is read at initialization time.

@par Compile-time dependencies

- none

@par Provides

- @ref interface_aio

@par Requires

- FTDI USB2Serial driver

@par Configuration requests

- none

@par Configuration file options

- serialPort (string)
  - Default: /dev/ttyUSB0
  - This defines the serial device where the glove is connected. Can be found out with dmesg after connecting

- provides
  - An "aio" interface shows the values that are coming from the glove sensors
  
@par Example 

@verbatim
driver
(
  name "dataglove"
  provides ["aio:0"]
  serialPort /dev/ttyUSB0
  alwayson 1
)
@endverbatim

@author Ingo Kresse

 */
/** @} */



#include <unistd.h>
#include <string.h>
//#include <iostream>
//#include <vector>
//#include <string>

#include <fcntl.h> // low-level serial commands
#include <termios.h> // serial commands

#include <libplayercore/playercore.h>


//For nanosleep:
#include <time.h>
#include <sys/time.h>
//To catch the errors of nanosleep
#include <errno.h>


//This function returns the difference in mS between two timeval structures
inline float timediffms(struct timeval start, struct timeval end) {
    return(end.tv_sec*1000.0 + end.tv_usec/1000.0 - (start.tv_sec*1000.0 + start.tv_usec/1000.0));
}



class Dataglove : public Driver {
public:

    // Constructor; need that
    Dataglove(ConfigFile* cf, int section);

    // Destructor
    ~Dataglove();

    // Must implement the following methods.
    virtual int Setup();
    virtual int Shutdown();

    // This method will be invoked on each incoming message
    virtual int ProcessMessage(QueuePointer &resp_queue,
                               player_msghdr * hdr,
                               void * data);

private:

    // Main function for device thread.
    virtual void Main();
	
	// config parameters
	const char *serialPort; // serial port device to use
	
	float alarmtime;
    float samplingrate;
	int fd; //port file handle
	struct termios oldSerialOptions;
	
	#define BUFFER_LEN 256
	char buffer[BUFFER_LEN+1];
	float raw[24]; //!< raw data values
	
	int firmware;
	char firmwareInfo[BUFFER_LEN+1],
	hardwareInfo[BUFFER_LEN+1],
	productInfo[BUFFER_LEN+1];
	
	//communication functions
	int readInfo(); //!< read version numbers
	int readValues(); //!< read all values from the device
	int parseValues(char *buffer); //!< parse the values
	
	int writeSerial(char cmd);
	int readSerial(char *buffer, int len);

    //! Player Interfaces
    player_devaddr_t aio_id;
};

// A factory creation function, declared outside of the class so that it
// can be invoked without any object context (alternatively, you can
// declare it static in the class).  In this function, we create and return
// (as a generic Driver*) a pointer to a new instance of this driver.
Driver*
Dataglove_Init(ConfigFile* cf, int section) {
    // Create and return a new instance of this driver
    return((Driver*)(new Dataglove(cf, section)));
}

// A driver registration function, again declared outside of the class so
// that it can be invoked without object context.  In this function, we add
// the driver into the given driver table, indicating which interface the
// driver can support and how to create a driver instance.
void Dataglove_Register(DriverTable* table) {
    table->AddDriver("dataglove", Dataglove_Init);
}

////////////////////////////////////////////////////////////////////////////////
// Constructor.  Retrieve options from the configuration file and do any
// pre-Setup() setup.
Dataglove::Dataglove(ConfigFile* cf, int section)
        : Driver(cf, section) {

    memset(&aio_id, 0, sizeof(player_devaddr_t));

    // Do we create an aio interface?
    if (cf->ReadDeviceAddr(&(aio_id), section, "provides",
                           PLAYER_AIO_CODE, -1, NULL) == 0) {
        if (AddInterface(aio_id) != 0) {
            SetError(-1);
            return;
        }
    } else {
        PLAYER_WARN("aio interface not created for Dataglove driver");
    }

// Read port name from the configuration file
    serialPort = cf->ReadString(section, "serialPort", "/dev/ttyUSB0");
}

//Destructor
Dataglove::~Dataglove() {}


////////////////////////////////////////////////////////////////////////////////
// Set up the device.  Return 0 if things go well, and -1 otherwise.
int Dataglove::Setup() {
    PLAYER_MSG0(1,"Dataglove driver initialising");

    // Here you do whatever is necessary to setup the device, like open and
    // configure a serial port.


	fd = -1;
	
	int rate = B1000000;
	
	// open the serial port
	
	//fd = open(port, O_RDWR | O_NOCTTY | O_NDELAY);
	fd = open(serialPort, O_RDWR | O_NOCTTY);
	if ( fd<0 )
	{
		fprintf(stderr, "Could not open serial device %s\n",serialPort);
		return -1;
	}
	
	
	// save the current io settings
	tcgetattr(fd, &oldSerialOptions);

	struct termios options;

	options=oldSerialOptions;
	
	// set data format to 8N1
	options.c_cflag &= ~(PARENB | CSTOPB | CSIZE);
	options.c_cflag |= CS8;	
	
	options.c_cflag &= ~CRTSCTS;       // disable hardware flow control
	options.c_cflag |= CLOCAL | CREAD; // process control stuff:
	                                   // don't get disturbed by special commands coming through the port
	                                   // allow reading from serial port

	options.c_lflag &= ~(ICANON | ECHO | ECHOE | ISIG); // raw input mode
	options.c_iflag &= ~(IXON | IXOFF | IXANY);         // disable software flow control
	
	options.c_oflag &= ~OPOST; // disable output postprocessing (= raw output)
		
	if (cfsetispeed(&options, rate) < 0 || cfsetospeed(&options, rate) < 0)
	{
		fprintf(stderr,"Failed to set serial baud rate: %d\n", rate);
		tcsetattr(fd, TCSANOW, &oldSerialOptions);	
		close(fd);
		fd = -1;
		return -1;
	}

	// activate new settings
	tcflush(fd, TCIFLUSH);
	tcsetattr(fd, TCSANOW, &options);
	tcflush(fd, TCIFLUSH);
	
	fcntl(fd, F_SETFL, FNDELAY); // read calls will NOT block during initialization
	
	int retries = 3;
	bool success=false;
	while(retries > 0 && !(success = (readInfo() >= 0))) {
		PLAYER_MSG0(1,"Dataglove: initialization failed, retrying...");
		retries--;
	} 
	
	if(!success) {
		// if it really failed give up and disable the driver
		PLAYER_MSG0(1,"Dataglove: initialization failed, giving up.");
		tcsetattr(fd, TCSANOW, &oldSerialOptions);
		close(fd);
		fd = -1;
		return -1;
	}
	
	fcntl(fd,F_SETFL, 0); // read calls will block in the main loop

    PLAYER_MSG0(1,"Dataglove driver ready");

    // Start the device thread; spawns a new thread and executes
    // Dataglove::Main(), which contains the main loop for the driver.
    StartThread();

    return(0);
}


////////////////////////////////////////////////////////////////////////////////
// Shutdown the device
int Dataglove::Shutdown() {
    PLAYER_MSG0(1,"Shutting Dataglove driver down");

    // Stop and join the driver thread
    StopThread();

    usleep(100000);

    PLAYER_MSG0(1,"Dataglove driver has been shutdown");

    return(0);
}

int Dataglove::ProcessMessage(QueuePointer &resp_queue,
                                     player_msghdr * hdr,
                                     void * data) {
    // Process messages here.  Send a response if necessary, using Publish().
    // If you handle the message successfully, return 0.  Otherwise,
    // return -1, and a NACK will be sent for you, if a response is required.

    return(-1); //return that the message wasn't handled
}



////////////////////////////////////////////////////////////////////////////////
// Main function for device thread
void Dataglove::Main() {

	player_aio_data_t data_ai;

	// The main loop; interact with the device here
    while (true) {

        // test if we are supposed to cancel
        pthread_testcancel();

        // Process incoming messages.  Dataglove::ProcessMessage() is
        // called on each message.
        ProcessMessages();

        //Read from the device
		readValues();
		int numValues = parseValues(buffer);

		//Construct the player data packet
		data_ai.voltages_count=numValues;
		data_ai.voltages = raw;

        //Time to publish the packets! (only for the interfaces defined in the config file)
        if (aio_id.interf !=0) {
            Publish(aio_id, PLAYER_MSGTYPE_DATA, PLAYER_AIO_DATA_STATE, (unsigned char*)&data_ai, sizeof(player_aio_data_t), NULL);
        }
    }
}

int Dataglove::readInfo()
{
	writeSerial(0x11); // get firmware version
	usleep(50000);
	int l1 = readSerial(firmwareInfo, 40);
	firmwareInfo[l1] = '\0'; // just to make shure
	
	// parse firmware version number
	firmware = (firmwareInfo[0]-'0')*100
			+ (firmwareInfo[2]-'0')*10
			+ (firmwareInfo[4]-'0');
	
	writeSerial(0x12); // get hardware version
	usleep(50000);
	int l2 = readSerial(hardwareInfo, 40);
	hardwareInfo[l2] = '\0'; // just to make shure
	
	writeSerial(0x13); // get hardware version
	usleep(50000);
	int l3 = readSerial(productInfo, 40);
	productInfo[l3] = '\0'; // just to make shure

	// delay here so data has arrived at serial port so we can flush it
	usleep(100000);
	tcflush(fd, TCIFLUSH);

	if (l1 < 3 || l2 < 3 || l3 < 3) // got an error
	{
		return -1;
	}
	
	fprintf(stderr, "Firmware Version: %s\nHardware Version: %s\nDevice Name: %s\n", firmwareInfo, hardwareInfo, productInfo);
	
	return 1;
}

// TODO: make this robust against lost bytes errors
// return 1 if successful, -1 otherwise
int Dataglove::readValues()
{
	struct timeval tStart;
	struct timeval tEnd;
	
	gettimeofday(&tStart, 0);

	if (fd < 0)
		return -1;
	
	writeSerial(0x35);
	usleep(10000);  // sleep a short while, cycle time is 16.6 ms
	int len = 0;
	bool retry=false;
	len = readSerial(buffer, 36-len);

	// TODO: add more error handling: flush buffers if everthing else fails
	while(len < 36)
	{
		retry = true;
		fprintf(stderr,"## Error getting data (got %d / 36), retrying...\n", len);
		len += readSerial(buffer+len, 36-len);
	}
	
	gettimeofday(&tEnd, 0);
	
	float tdiff = tEnd.tv_sec*1000.0 + tEnd.tv_usec/1000.0 - tStart.tv_sec*1000.0 - tStart.tv_usec/1000.0;
	
	if(tdiff > 25.0) {
		fprintf(stderr, "## overlong cyle: %f ms\n", tdiff);
	}
	
	if(retry)
		fprintf(stderr, "## ... packet complete.\n");
	
	return 1;
}

// parse the values in buffer, and store the values in raw[]
// returns the number of values it converted
int Dataglove::parseValues(char *buffer)
{
	unsigned char *buf;
	buf = (unsigned char *) buffer;

	raw[0] = buf[0] + (buf[1]/16)*256;
	raw[1] = buf[2] + (buf[1]%16)*256;
	raw[2] = buf[3] + (buf[4]/16)*256;
	raw[3] = buf[5] + (buf[4]%16)*256;
	raw[4] = buf[6] + (buf[7]/16)*256;
	raw[5] = buf[8] + (buf[7]%16)*256;
	raw[6] = buf[9] + (buf[10]/16)*256;
	raw[7] = buf[11] + (buf[10]%16)*256;
	raw[8] = buf[12] + (buf[13]/16)*256;
	raw[9] = buf[14] + (buf[13]%16)*256;
	raw[10] = buf[15] + (buf[16]/16)*256;
	raw[11] = buf[17] + (buf[16]%16)*256;
	raw[12] = buf[18] + (buf[19]/16)*256;
	raw[13] = buf[20] + (buf[19]%16)*256;
	raw[14] = buf[21] + (buf[22]/16)*256;
	raw[15] = buf[23] + (buf[22]%16)*256;
	raw[16] = buf[24] + (buf[25]/16)*256;
	raw[17] = buf[26] + (buf[25]%16)*256;
	raw[18] = buf[27] + (buf[28]/16)*256;
	raw[19] = buf[29] + (buf[28]%16)*256;
	if (firmware > 26)
	{
		raw[20] = buf[30] + (buf[31]/16)*256;
		raw[21] = buf[32] + (buf[31]%16)*256;
		raw[22] = buf[33] + (buf[34]/16)*256;
		raw[23] = buf[35] + (buf[34]%16)*256;
		return 24;
	}
	else
		return 20;
}



int Dataglove::writeSerial(char cmd)
{
	if (fd < 0)
		return -1;
	if(write(fd, &cmd, 1) < 1)
		return -1;
//  fprintf(stderr, "## written 0x%02x\n",cmd);

	return 1;
}


int Dataglove::readSerial(char *buffer, int len)
{
	int l;
	if (fd < 0)
		return -1;
	l=read(fd, buffer, len );
//	fprintf(stderr, "## read %d chars\n", l);
	return l;
}


extern "C" {
  int player_driver_init(DriverTable* table)
  {
    puts("Dataglove driver initializing");
    Dataglove_Register(table);
    puts("Dataglove driver done");
    return(0);
  }
}
